@@ -1,10 +1,11 @@
 /*
- * Copyright (C) 2001 Sistina Software (UK) Limited.
+ * Copyright (C) 2001-2003 Sistina Software (UK) Limited.
  *
  * This file is released under the GPL.
  */
 
 #include "libdevmapper.h"
+#include "log.h"
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -269,7 +270,7 @@ static int _version(int argc, char **argv, void *data)
 	return r;
 }
 
-static int _simple(int task, const char *name, int display)
+static int _simple(int task, const char *name, uint32_t event_nr, int display)
 {
 	int r = 0;
 
@@ -282,6 +283,9 @@ static int _simple(int task, const char *name, int display)
 	if (!dm_task_set_name(dmt, name))
 		goto out;
 
+	if (event_nr && !dm_task_set_event_nr(dmt, event_nr))
+		goto out;
+
 	r = dm_task_run(dmt);
 
 	if (r && display && _switches[VERBOSE_ARG])
@@ -294,32 +298,33 @@ static int _simple(int task, const char *name, int display)
 
 static int _remove_all(int argc, char **argv, void *data)
 {
-	return _simple(DM_DEVICE_REMOVE_ALL, "", 0);
+	return _simple(DM_DEVICE_REMOVE_ALL, "", 0, 0);
 }
 
 static int _remove(int argc, char **argv, void *data)
 {
-	return _simple(DM_DEVICE_REMOVE, argv[1], 0);
+	return _simple(DM_DEVICE_REMOVE, argv[1], 0, 0);
 }
 
 static int _suspend(int argc, char **argv, void *data)
 {
-	return _simple(DM_DEVICE_SUSPEND, argv[1], 1);
+	return _simple(DM_DEVICE_SUSPEND, argv[1], 0, 1);
 }
 
 static int _resume(int argc, char **argv, void *data)
 {
-	return _simple(DM_DEVICE_RESUME, argv[1], 1);
+	return _simple(DM_DEVICE_RESUME, argv[1], 0, 1);
 }
 
 static int _clear(int argc, char **argv, void *data)
 {
-	return _simple(DM_DEVICE_CLEAR, argv[1], 1);
+	return _simple(DM_DEVICE_CLEAR, argv[1], 0, 1);
 }
 
 static int _wait(int argc, char **argv, void *data)
 {
-	return _simple(DM_DEVICE_WAITEVENT, argv[1], 2);
+	return _simple(DM_DEVICE_WAITEVENT, argv[1],
+		       (argc == 3) ? atoi(argv[2]) : 0, 1);
 }
 
 static int _process_mapper_dir(int argc, char **argv,
@@ -458,6 +463,40 @@ static int _status(int argc, char **argv, void *data)
 
 }
 
+/* Show target names and their version numbers */
+static int _targets(int argc, char **argv, void *data)
+{
+	int r = 0;
+	struct dm_task *dmt;
+	struct dm_versions *target;
+	struct dm_versions *last_target;
+
+	if (!(dmt = dm_task_create(DM_DEVICE_LIST_VERSIONS)))
+		return 0;
+
+	if (!dm_task_run(dmt))
+		goto out;
+
+	target = dm_task_get_versions(dmt);
+
+	/* Fetch targets and print 'em */
+	do {
+		last_target = target;
+
+		printf("%-16s v%d.%d.%d\n", target->name, target->version[0],
+		       target->version[1], target->version[2]);
+
+		target = (void *) target + target->next;
+	} while (last_target != target);
+
+	r = 1;
+
+      out:
+	dm_task_destroy(dmt);
+	return r;
+
+}
+
 static int _info(int argc, char **argv, void *data)
 {
 	int r = 0;
@@ -606,7 +645,8 @@ static struct command _commands[] = {
 	{"mknodes", "[<dev_name>]", 0, 1, _info},
 	{"status", "[<dev_name>]", 0, 1, _status},
 	{"table", "[<dev_name>]", 0, 1, _status},
-	{"wait", "<dev_name>", 1, 1, _wait},
+	{"wait", "<dev_name> [<event_nr>]", 1, 2, _wait},
+	{"targets", "", 0, 0, _targets},
 	{"version", "", 0, 0, _version},
 	{NULL, NULL, 0, 0, NULL}
 };
