@@ -6,7 +6,11 @@
 
 #include "libdm-targets.h"
 #include "libdm-common.h"
-#include "libdm-compat.h"
+#include "log.h"
+
+#ifdef DM_COMPAT
+#  include "libdm-compat.h"
+#endif
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -16,8 +20,18 @@
 #include <dirent.h>
 #include <errno.h>
 #include <sys/stat.h>
+#include <sys/ioctl.h>
 #include <limits.h>
-#include <linux/limits.h>
+
+#ifdef linux
+#  include <linux/limits.h>
+#  include <linux/kdev_t.h>
+#  include <linux/dm-ioctl.h>
+#else
+#  define MAJOR(x) major((x))
+#  define MINOR(x) minor((x))
+#  define MKDEV(x,y) makedev((x),(y))
+#endif
 
 /*
  * Ensure build compatibility.  
@@ -66,6 +80,8 @@ static struct cmd_data _cmd_data_v4[] = {
 	{"waitevent",	DM_DEV_WAIT,		{4, 0, 0}},
 	{"names",	DM_LIST_DEVICES,	{4, 0, 0}},
 	{"clear",	DM_TABLE_CLEAR,		{4, 0, 0}},
+	{"mknodes",	DM_DEV_STATUS,		{4, 0, 0}},
+	{"versions",	DM_LIST_VERSIONS,	{4, 1, 0}},
 };
 /* *INDENT-ON* */
 
@@ -84,6 +100,56 @@ static void *_align(void *ptr, unsigned int a)
 	return (void *) (((unsigned long) ptr + agn) & ~agn);
 }
 
+static int _open_control(void)
+{
+	char control[PATH_MAX];
+
+	if (_control_fd != -1)
+		return 1;
+
+	snprintf(control, sizeof(control), "%s/control", dm_dir());
+
+	if ((_control_fd = open(control, O_RDWR)) < 0) {
+		log_error("%s: open failed: %s", control, strerror(errno));
+		log_error("Is device-mapper driver missing from kernel?");
+		return 0;
+	}
+
+	return 1;
+}
+
+void dm_task_destroy(struct dm_task *dmt)
+{
+	struct target *t, *n;
+
+	for (t = dmt->head; t; t = n) {
+		n = t->next;
+		free(t->params);
+		free(t->type);
+		free(t);
+	}
+
+	if (dmt->dev_name)
+		free(dmt->dev_name);
+
+	if (dmt->newname)
+		free(dmt->newname);
+
+	if (dmt->dmi.v4)
+		free(dmt->dmi.v4);
+
+	if (dmt->uuid)
+		free(dmt->uuid);
+
+	free(dmt);
+}
+
+/*
+ * Protocol Version 1 compatibility functions.
+ */
+
+#ifdef DM_COMPAT
+
 static int _dm_task_get_driver_version_v1(struct dm_task *dmt, char *version,
 					  size_t size)
 {
@@ -122,7 +188,8 @@ static int _unmarshal_status_v1(struct dm_task *dmt, struct dm_ioctl_v1 *dmi)
 	return 1;
 }
 
-static int _dm_format_dev_v1(char *buf, int bufsize, uint32_t dev_major, uint32_t dev_minor)
+static int _dm_format_dev_v1(char *buf, int bufsize, uint32_t dev_major,
+			     uint32_t dev_minor)
 {
 	int r;
 
@@ -319,6 +386,10 @@ static int _dm_names_v1(struct dm_ioctl_v1 *dmi)
 	struct stat buf;
 	char path[PATH_MAX];
 
+	log_print("Warning: Device list may be incomplete with interface "
+		  "version 1.");
+	log_print("Please upgrade your kernel device-mapper driver.");
+
 	if (!(d = opendir(dev_dir))) {
 		log_error("%s: opendir failed: %s", dev_dir, strerror(errno));
 		return 0;
@@ -365,24 +436,6 @@ static int _dm_names_v1(struct dm_ioctl_v1 *dmi)
 	return r;
 }
 
-static int _open_control(void)
-{
-	char control[PATH_MAX];
-
-	if (_control_fd != -1)
-		return 1;
-
-	snprintf(control, sizeof(control), "%s/control", dm_dir());
-
-	if ((_control_fd = open(control, O_RDWR)) < 0) {
-		log_error("%s: open failed: %s", control, strerror(errno));
-		log_error("Is device-mapper driver missing from kernel?");
-		return 0;
-	}
-
-	return 1;
-}
-
 static int _dm_task_run_v1(struct dm_task *dmt)
 {
 	struct dm_ioctl_v1 *dmi;
@@ -437,6 +490,14 @@ static int _dm_task_run_v1(struct dm_task *dmt)
 		rename_dev_node(dmt->dev_name, dmt->newname);
 		break;
 
+	case DM_DEVICE_MKNODES:
+		if (dmi->flags & DM_EXISTS_FLAG)
+			add_dev_node(dmt->dev_name, MAJOR(dmi->dev),
+				     MINOR(dmi->dev));
+		else
+			rm_dev_node(dmt->dev_name);
+		break;
+
 	case DM_DEVICE_STATUS:
 	case DM_DEVICE_TABLE:
 		if (!_unmarshal_status_v1(dmt, dmi))
@@ -460,38 +521,20 @@ static int _dm_task_run_v1(struct dm_task *dmt)
 	return 0;
 }
 
-void dm_task_destroy(struct dm_task *dmt)
-{
-	struct target *t, *n;
+#endif
 
-	for (t = dmt->head; t; t = n) {
-		n = t->next;
-		free(t->params);
-		free(t->type);
-		free(t);
-	}
-
-	if (dmt->dev_name)
-		free(dmt->dev_name);
-
-	if (dmt->newname)
-		free(dmt->newname);
-
-	if (dmt->dmi.v4)
-		free(dmt->dmi.v4);
-
-	if (dmt->uuid)
-		free(dmt->uuid);
-
-	free(dmt);
-}
+/*
+ * Protocol Version 4 functions.
+ */
 
 int dm_task_get_driver_version(struct dm_task *dmt, char *version, size_t size)
 {
 	unsigned int *v;
 
+#ifdef DM_COMPAT
 	if (_dm_version == 1)
 		return _dm_task_get_driver_version_v1(dmt, version, size);
+#endif
 
 	if (!dmt->dmi.v4) {
 		version[0] = '\0';
@@ -607,12 +650,15 @@ static int _unmarshal_status(struct dm_task *dmt, struct dm_ioctl *dmi)
 	return 1;
 }
 
-int dm_format_dev(char *buf, int bufsize, uint32_t dev_major, uint32_t dev_minor)
+int dm_format_dev(char *buf, int bufsize, uint32_t dev_major,
+		  uint32_t dev_minor)
 {
 	int r;
 
+#ifdef DM_COMPAT
 	if (_dm_version == 1)
 		return _dm_format_dev_v1(buf, bufsize, dev_major, dev_minor);
+#endif
 
 	if (bufsize < 8)
 		return 0;
@@ -626,8 +672,10 @@ int dm_format_dev(char *buf, int bufsize, uint32_t dev_major, uint32_t dev_minor
 
 int dm_task_get_info(struct dm_task *dmt, struct dm_info *info)
 {
+#ifdef DM_COMPAT
 	if (_dm_version == 1)
 		return _dm_task_get_info_v1(dmt, info);
+#endif
 
 	if (!dmt->dmi.v4)
 		return 0;
@@ -654,24 +702,30 @@ int dm_task_get_info(struct dm_task *dmt, struct dm_info *info)
 
 const char *dm_task_get_name(struct dm_task *dmt)
 {
+#ifdef DM_COMPAT
 	if (_dm_version == 1)
 		return _dm_task_get_name_v1(dmt);
+#endif
 
 	return (dmt->dmi.v4->name);
 }
 
 const char *dm_task_get_uuid(struct dm_task *dmt)
 {
+#ifdef DM_COMPAT
 	if (_dm_version == 1)
 		return _dm_task_get_uuid_v1(dmt);
+#endif
 
 	return (dmt->dmi.v4->uuid);
 }
 
 struct dm_deps *dm_task_get_deps(struct dm_task *dmt)
 {
+#ifdef DM_COMPAT
 	if (_dm_version == 1)
 		return _dm_task_get_deps_v1(dmt);
+#endif
 
 	return (struct dm_deps *) (((void *) dmt->dmi.v4) +
 				   dmt->dmi.v4->data_start);
@@ -679,13 +733,21 @@ struct dm_deps *dm_task_get_deps(struct dm_task *dmt)
 
 struct dm_names *dm_task_get_names(struct dm_task *dmt)
 {
+#ifdef DM_COMPAT
 	if (_dm_version == 1)
 		return _dm_task_get_names_v1(dmt);
+#endif
 
 	return (struct dm_names *) (((void *) dmt->dmi.v4) +
 				    dmt->dmi.v4->data_start);
 }
 
+struct dm_versions *dm_task_get_versions(struct dm_task *dmt)
+{
+	return (struct dm_versions *) (((void *) dmt->dmi.v4) +
+				       dmt->dmi.v4->data_start);
+}
+
 int dm_task_set_ro(struct dm_task *dmt)
 {
 	dmt->read_only = 1;
@@ -933,11 +995,13 @@ static int _create_and_load_v4(struct dm_task *dmt)
 
 int dm_task_run(struct dm_task *dmt)
 {
-	struct dm_ioctl *dmi;
+	struct dm_ioctl *dmi = NULL;
 	unsigned int command;
 
+#ifdef DM_COMPAT
 	if (_dm_version == 1)
 		return _dm_task_run_v1(dmt);
+#endif
 
 	if ((unsigned) dmt->type >=
 	    (sizeof(_cmd_data_v4) / sizeof(*_cmd_data_v4))) {
@@ -968,7 +1032,8 @@ int dm_task_run(struct dm_task *dmt)
 	log_debug("dm %s %s %s %s", _cmd_data_v4[dmt->type].name, dmi->name,
 		  dmi->uuid, dmt->newname ? dmt->newname : "");
 	if (ioctl(_control_fd, command, dmi) < 0) {
-		if (errno == ENXIO && dmt->type == DM_DEVICE_INFO) {
+		if (errno == ENXIO && ((dmt->type == DM_DEVICE_INFO) ||
+				       (dmt->type == DM_DEVICE_MKNODES))) {
 			dmi->flags &= ~DM_EXISTS_FLAG;	/* FIXME */
 			goto ignore_error;
 		}
@@ -995,6 +1060,14 @@ int dm_task_run(struct dm_task *dmt)
 		rename_dev_node(dmt->dev_name, dmt->newname);
 		break;
 
+	case DM_DEVICE_MKNODES:
+		if (dmi->flags & DM_EXISTS_FLAG)
+			add_dev_node(dmt->dev_name, MAJOR(dmi->dev),
+				     MINOR(dmi->dev));
+		else
+			rm_dev_node(dmt->dev_name);
+		break;
+
 	case DM_DEVICE_STATUS:
 	case DM_DEVICE_TABLE:
 	case DM_DEVICE_WAITEVENT:
