@@ -9,7 +9,10 @@
 
 #include <inttypes.h>
 #include <sys/types.h>
-#include <linux/types.h>
+
+#ifdef linux
+#  include <linux/types.h>
+#endif
 
 /*
  * Since it is quite laborious to build the ioctl
@@ -25,8 +28,8 @@ typedef void (*dm_log_fn) (int level, const char *file, int line,
 
 /*
  * The library user may wish to register their own
- * logging function, by default errors go to
- * stderr.
+ * logging function, by default errors go to stderr.
+ * Use dm_log_init(NULL) to restore the default log fn.
  */
 void dm_log_init(dm_log_fn fn);
 void dm_log_init_verbose(int level);
@@ -48,7 +51,15 @@ enum {
 
 	DM_DEVICE_STATUS,
 	DM_DEVICE_TABLE,
-	DM_DEVICE_WAITEVENT
+	DM_DEVICE_WAITEVENT,
+
+	DM_DEVICE_LIST,
+
+	DM_DEVICE_CLEAR,
+
+	DM_DEVICE_MKNODES,
+
+	DM_DEVICE_LIST_VERSIONS
 };
 
 struct dm_task;
@@ -65,29 +76,51 @@ int dm_task_set_uuid(struct dm_task *dmt, const char *uuid);
 struct dm_info {
 	int exists;
 	int suspended;
-	unsigned int open_count;
-	int major;
-	int minor;		/* minor device number */
+	int live_table;
+	int inactive_table;
+	int32_t open_count;
+	uint32_t event_nr;
+	uint32_t major;
+	uint32_t minor;		/* minor device number */
 	int read_only;		/* 0:read-write; 1:read-only */
 
-	unsigned int target_count;
+	int32_t target_count;
 };
 
 struct dm_deps {
-	unsigned int count;
-	__kernel_dev_t device[0];
+	uint32_t count;
+	uint32_t filler;
+	uint64_t device[0];
+};
+
+struct dm_names {
+	uint64_t dev;
+	uint32_t next;		/* Offset to next struct from start of this struct */
+	char name[0];
+};
+
+struct dm_versions {
+        uint32_t next;		/* Offset to next struct from start of this struct */
+        uint32_t version[3];
+
+        char name[0];
 };
 
 int dm_get_library_version(char *version, size_t size);
 int dm_task_get_driver_version(struct dm_task *dmt, char *version, size_t size);
 int dm_task_get_info(struct dm_task *dmt, struct dm_info *dmi);
+const char *dm_task_get_name(struct dm_task *dmt);
 const char *dm_task_get_uuid(struct dm_task *dmt);
 
 struct dm_deps *dm_task_get_deps(struct dm_task *dmt);
+struct dm_names *dm_task_get_names(struct dm_task *dmt);
+struct dm_versions *dm_task_get_versions(struct dm_task *dmt);
 
 int dm_task_set_ro(struct dm_task *dmt);
 int dm_task_set_newname(struct dm_task *dmt, const char *newname);
 int dm_task_set_minor(struct dm_task *dmt, int minor);
+int dm_task_set_major(struct dm_task *dmt, int major);
+int dm_task_set_event_nr(struct dm_task *dmt, uint32_t event_nr);
 
 /*
  * Use these to prepare for a create or reload.
@@ -96,6 +129,11 @@ int dm_task_add_target(struct dm_task *dmt,
 		       uint64_t start,
 		       uint64_t size, const char *ttype, const char *params);
 
+/*
+ * Format major/minor numbers correctly for input to driver
+ */
+int dm_format_dev(char *buf, int bufsize, uint32_t dev_major, uint32_t dev_minor);
+
 /* Use this to retrive target information returned from a STATUS call */
 void *dm_get_next_target(struct dm_task *dmt,
 			 void *next, uint64_t *start, uint64_t *length,
@@ -112,4 +150,8 @@ int dm_task_run(struct dm_task *dmt);
 int dm_set_dev_dir(const char *dir);
 const char *dm_dir(void);
 
+/* Release library resources */
+void dm_lib_release(void);
+void dm_lib_exit(void);
+
 #endif				/* LIB_DEVICE_MAPPER_H */
